package dhm 
{
	import com.efg.framework.BlitSprite;
	import com.efg.framework.CustomEventLevelScreenUpdate;
	import com.efg.framework.TileSheet;
	import flash.display.*;
	import flash.filters.*;
	import flash.geom.*;
	
	/**
	 * ...
	 * @author DHM
	 * 
	 * Base class for Player and Enemy classes.
	 */
	public class Contestant extends MyBlitSprite
	{
		private static const DEATH_FADE_RATE:Number = -0.03;
		private static const HIT_FLASH_DURATION:int = Main.FRAME_RATE / 4;
		
		private static const FACING_RIGHT_TILE_LIST:Array = [0, 1];
		private static const FACING_LEFT_TILE_LIST:Array = [2, 3];
		private static const FACING_UP_TILE_LIST:Array = [4, 5, 6, 5];
		private static const FACING_DOWN_TILE_LIST:Array = [7, 8, 9, 8];
		
		public static const WIDTH:int = 32;
		public static const HEIGHT:int = 32;
		
		private static const DIAGONAL_SPEED_FACTOR:Number = 1 / Math.sqrt(2);
			
		private var _health:int; // number of hit points remaining
		
		private var _speed:Number;
		private var _boostCounter:int; // number of frames remaining on speed boost
		private var _boostFactor:Number = 1; // speed boost multiplier
		
		private var _facing:int = Facing.RIGHT;
		private var _velocity:Point = new Point(0, 0);
		private var _nextPosition:Point = new Point();
		
		private var _weapon:Weapon;
		
		// hit flash
		private var _flashCounter:int;
		private var _flashFilter:GlowFilter = new GlowFilter(0xCCCCCC, 0.6, 8, 8, 2, 1, true);
		
		// invincibility
		private var _invincibleCounter:int;
		private var _invincibleFilter:GlowFilter = new GlowFilter(0xDDDDDD, 0.6, 8, 8, 3, 1, false);
		
		protected var _positionBounds:Rectangle; // upper/lower bounds on position

		public var hitTestExclusions:Array = []; // keep track of which objects to ignore when hit testing
		
		public function Contestant(tileImage:Bitmap, worldBounds:Rectangle, weapon:Weapon, speed:Number, health:int)
		{
			var tileSheet:TileSheet = new TileSheet(tileImage.bitmapData, WIDTH, HEIGHT);
			
			super(tileSheet, FACING_RIGHT_TILE_LIST, 0);
			
			_speed = speed;
			_health = health;
			_weapon = weapon;
			
			//graphics.clear();
			//graphics.beginFill(color);
			//graphics.drawRoundRect(0, 0, WIDTH, HEIGHT, 4, 4);
			//graphics.endFill();
			
			_positionBounds = new Rectangle(worldBounds.left, worldBounds.top, worldBounds.width - WIDTH, worldBounds.height - HEIGHT);
			
			onFacingChanged();
		}
		
		public function get isAlive():Boolean { return _health > 0; }
		public function get shouldRemove():Boolean { return alpha <= 0; }
		public function get weapon():Weapon { return _weapon; }
		
		public function get center():Point { return new Point(x + WIDTH / 2, y + HEIGHT / 2); }
		public function get speed():Number { return _speed; }
		public function get velocity():Point { return _velocity; }
		
		public function get facing():int { return _facing; }
		public function set facing(value:int):void
		{
			if (value == _facing)
				return;
			_facing = value;
			onFacingChanged();
		}
		
		public function get health():int { return _health; }
		public function set health(value:int):void
		{
			_health = Math.max(0, Math.min(Settings.MAXIMUM_CONTESTANT_HEALTH, value));
		}
		
		public function reset():void
		{
			facing = Facing.RIGHT;
			_boostFactor = 1;
			_boostCounter = 0; 
			_invincibleCounter = 0;
			updateFilters();
			weapon.reset();
			
			update();
			render();
		}
		
		public function hit(damage:int):void
		{
			_health = Math.max(0, _health - damage);
			_flashCounter = HIT_FLASH_DURATION;
			updateFilters();
		}
		
		public function makeInvincible(frameCount:int):void
		{
			_invincibleCounter = frameCount;
			updateFilters();
			hitTestExclusions.splice(0, hitTestExclusions.length);
		}
		
		public function get isInvincible():Boolean { return _invincibleCounter > 0; }
		
		// apply a boost factor (speed multiplier) for a certain number of frames
		public function boostSpeed(factor:int, frameCount:int):void
		{
			_boostFactor = factor;
			_boostCounter = frameCount;
			_weapon.boostFactor = factor;
		}
		
		public function update():void
		{
			if (!isAlive)
				return;
			
			_nextPosition.x = x + _velocity.x;
			_nextPosition.y = y + _velocity.y;
			
			// disallow movement outside of world boundaries
			constrainPosition(_nextPosition);
			
			super.updateCurrentTile();
			
			// pull the weapon along with its owner
			_weapon.updateHomePosition(getWeaponHomePosition(_nextPosition));
			_weapon.update();
		}
		
		public function render():void
		{
			updateFlashCounter();
			
			if (isAlive)
			{
				x = _nextPosition.x;
				y = _nextPosition.y;
				
				super.renderCurrentTile();
				
				_weapon.render();
					
				updateBoostCounter();
				updateInvincibleCounter();
			}
			else
			{
				_weapon.alpha += DEATH_FADE_RATE;
				alpha += DEATH_FADE_RATE;
			}
		}
		
		public function moveTo(position:Point):void
		{
			_nextPosition.x = x = position.x;
			_nextPosition.y = y = position.y;
			_weapon.moveTo(getWeaponHomePosition(position));
		}
		
		protected function constrainPosition(p:Point):void
		{
			p.x = Math.min(Math.max(_positionBounds.left, p.x), _positionBounds.right);
			p.y = Math.min(Math.max(_positionBounds.top, p.y), _positionBounds.bottom);
		}
		
		// Movement is constrained to the horizontal and vertical axes, and the two
		// corresponding diagonals.  The dx and dy values are either positive, negative, or zero.
		// Example 1: dx == 0 && dy < 0 => move straight up
		// Example 2: dx < 0 && dy > 0 => move down and to the left on a diagonal
		protected function updateDirection(dx:int, dy:int):void
		{
			if (dx == 0 && dy == 0)
			{
				animationLoop = false;
				_velocity.x = _velocity.y = 0;
				return;
			}
			
			animationLoop = true;
			
			var effectiveSpeed:Number = _speed * _boostFactor;
			
			var componentSpeed:Number = (dx == 0 || dy == 0) ? effectiveSpeed : effectiveSpeed * DIAGONAL_SPEED_FACTOR;
			
			_velocity.x = dx == 0 ? 0 : dx < 0 ? -componentSpeed : componentSpeed;
			_velocity.y = dy == 0 ? 0 : dy < 0 ? -componentSpeed : componentSpeed;
			
			onVelocityChanged();
		}
		
		protected function onVelocityChanged():void
		{
			if (_velocity.x == 0 && _velocity.y == 0)
				return; // no change in direction
			
			if(Math.abs(_velocity.x) >= Math.abs(_velocity.y))
				facing = _velocity.x > 0 ? Facing.RIGHT : Facing.LEFT;
			else
				facing = _velocity.y > 0 ? Facing.DOWN : Facing.UP;
		}
		
		private function onFacingChanged():void
		{
			switch(_facing)
			{
				case Facing.RIGHT:
					updateTileList(FACING_RIGHT_TILE_LIST);
					break;
				case Facing.LEFT:
					updateTileList(FACING_LEFT_TILE_LIST);
					break;
				case Facing.UP:
					updateTileList(FACING_UP_TILE_LIST);
					break;
				case Facing.DOWN:
					updateTileList(FACING_DOWN_TILE_LIST);
					break;
			}
			
			_weapon.facing = facing;
		}
		
		// Offset the weapon to look like it is being held in the contestant's hand.
		// This, of course, assumes knowledge of the bitmap image, and would need to be 
		// overridden if an extension class used different images.  However, all of our
		// contestant characters use the same bitmap frames -- only the colors vary.
		protected function getWeaponHomePosition(relativeTo:Point):Point
		{
			var p:Point = new Point(relativeTo.x, relativeTo.y);
			
			switch(_facing)
			{
			case Facing.RIGHT:
				p.y += (HEIGHT / 2) - (_weapon.height / 2);
				p.x += WIDTH - 2;
				break;
			case Facing.LEFT:
				p.y += (HEIGHT / 2) - (_weapon.height / 2);
				p.x += - _weapon.width + 2;
				break;
			case Facing.DOWN:
				p.y += (HEIGHT / 2) + (_weapon.height / 2);
				p.x += 6;
				break;
			case Facing.UP:
				p.y += -_weapon.height;
				p.x += WIDTH - _weapon.width - 4;
				break;
			}
			return p;
		}
		
		private function updateFlashCounter():void 
		{
			if (_flashCounter > 0)
			{
				_flashCounter--;
				if (_flashCounter <= 0)
					updateFilters();
			}
		}
		
		private function updateBoostCounter():void
		{
			if (_boostCounter > 0)
			{
				_boostCounter--;
				if (_boostCounter <= 0)
				{
					_boostFactor = 1;
					_weapon.boostFactor = 1;
				}
			}
		}
		
		private function updateInvincibleCounter():void 
		{
			if (_invincibleCounter > 0)
			{
				_invincibleCounter--;
				if (_invincibleCounter <= 0)						
					updateFilters();
			}
		}
		
		private function updateFilters():void
		{
			// Apparently there is no collection change listener on the filters array.
			// Modifying the existing filters array and passing it to the filters setter does not cause the filters to render either.
			// The only way to get the filters to render (that I could find) is to call the setter with a new array of filters.
			var filters:Array = [];
			if (_invincibleCounter > 0)
				filters.push(_invincibleFilter);
			if (_flashCounter > 0)
				filters.push(_flashFilter);
			this.filters = filters;
		}
	}
}